from __future__ import annotations
import ast
import re
from typing import TYPE_CHECKING, Optional
from attr import define, field
from jsonschema.exceptions import ValidationError
from jsonschema.validators import validate
from skatepark.artifacts import TextOutput, ErrorOutput
from skatepark.steps import PromptStep
from griptape.core import BaseTool
from skatepark.utils import J2

if TYPE_CHECKING:
    from skatepark.artifacts import StructureArtifact
    from skatepark.steps import ToolkitStep


@define
class ToolSubstep(PromptStep):
    THOUGHT_PATTERN = r"^Thought:\s*(.*)$"
    ACTION_PATTERN = r"^Action:\s*({.*})$"
    OUTPUT_PATTERN = r"^Output:\s?([\s\S]*)$"
    INVALID_ACTION_ERROR_MSG = f"invalid action input, try again"

    parent_step_id: Optional[str] = field(default=None, kw_only=True)
    thought: Optional[str] = field(default=None, kw_only=True)
    tool_name: Optional[str] = field(default=None, kw_only=True)
    tool_action: Optional[str] = field(default=None, kw_only=True)
    tool_value: Optional[str] = field(default=None, kw_only=True)

    _tool: Optional[BaseTool] = None

    def attach(self, parent_step: ToolkitStep):
        self.parent_step_id = parent_step.id
        self.structure = parent_step.structure
        self.__init_from_prompt(self.render_prompt())

    @property
    def toolkit_step(self) -> Optional[ToolkitStep]:
        return self.structure.find_step(self.parent_step_id)

    @property
    def parents(self) -> list[ToolSubstep]:
        return [self.toolkit_step.find_substep(parent_id) for parent_id in self.parent_ids]

    @property
    def children(self) -> list[ToolSubstep]:
        return [self.toolkit_step.find_substep(child_id) for child_id in self.child_ids]

    def before_run(self) -> None:
        self.structure.logger.info(f"Substep {self.id}\n{self.render_prompt()}")

    def run(self) -> StructureArtifact:
        try:
            if self.tool_name == "error":
                self.output = ErrorOutput(self.tool_value, step=self)
            else:
                if self._tool:
                    observation = self.structure.tool_loader.executor.execute(
                        getattr(self._tool, self.tool_action),
                        self.tool_value.encode()
                    ).decode()
                else:
                    observation = "tool not found"

                self.output = TextOutput(observation)
        except Exception as e:
            self.structure.logger.error(f"Substep {self.id}\n{e}", exc_info=True)

            self.output = ErrorOutput(str(e), exception=e, step=self)
        finally:
            return self.output

    def after_run(self) -> None:
        self.structure.logger.info(f"Substep {self.id}\nObservation: {self.output.value}")

    def render(self) -> str:
        return J2("prompts/steps/tool/substep.j2").render(
            substep=self
        )

    def add_child(self, child: ToolSubstep) -> ToolSubstep:
        if child.id not in self.child_ids:
            self.child_ids.append(child.id)

        if self.id not in child.parent_ids:
            child.parent_ids.append(self.id)

        return child

    def add_parent(self, parent: ToolSubstep) -> ToolSubstep:
        if parent.id not in self.parent_ids:
            self.parent_ids.append(parent.id)

        if self.id not in parent.child_ids:
            parent.child_ids.append(self.id)

        return parent

    def __init_from_prompt(self, value: str) -> None:
        thought_matches = re.findall(self.THOUGHT_PATTERN, value, re.MULTILINE)
        action_matches = re.findall(self.ACTION_PATTERN, value, re.MULTILINE)
        output_matches = re.findall(self.OUTPUT_PATTERN, value, re.MULTILINE)

        if self.thought is None and len(thought_matches) > 0:
            self.thought = thought_matches[-1]

        if len(action_matches) > 0:
            try:
                parsed_value = ast.literal_eval(action_matches[-1])

                # Load the tool name; throw exception if the key is not present
                if self.tool_name is None:
                    self.tool_name = parsed_value["tool"]

                # Load the tool action; throw exception if the key is not present
                if self.tool_action is None:
                    self.tool_action = parsed_value["action"]

                # Load the tool itself
                if self.tool_name:
                    self._tool = self.toolkit_step.find_tool(self.tool_name)

                # Validate input based on tool schema
                if self._tool:
                    validate(
                        instance={
                            "value": parsed_value["value"]
                        },
                        schema=self._tool.action_schema(getattr(self._tool, self.tool_action))
                    )

                # Load optional input value; don't throw exceptions if key is not present
                if self.tool_value is None:
                    self.tool_value = str(parsed_value.get("value"))

            except SyntaxError as e:
                self.structure.logger.error(f"Step {self.toolkit_step.id}\nSyntax error: {e}")

                self.tool_name = "error"
                self.tool_value = f"syntax error: {e}"
            except ValidationError as e:
                self.structure.logger.error(f"Step {self.toolkit_step.id}\nInvalid JSON: {e}")

                self.tool_name = "error"
                self.tool_value = f"JSON validation error: {e}"
            except Exception as e:
                self.structure.logger.error(f"Step {self.toolkit_step.id}\nError parsing tool action: {e}")

                self.tool_name = "error"
                self.tool_value = f"error: {self.INVALID_ACTION_ERROR_MSG}"

        if self.output is None and len(output_matches) > 0:
            self.output = TextOutput(output_matches[-1])



from abc import abstractmethod
from marshmallow import fields
from marshmallow_enum import EnumField
from skatepark.schemas import BaseSchema
from skatepark.steps import Step


class StepSchema(BaseSchema):
    class Meta:
        ordered = True

    id = fields.Str()
    state = EnumField(Step.State)
    parent_ids = fields.List(fields.Str())
    child_ids = fields.List(fields.Str())

    @abstractmethod
    def make_obj(self, data, **kwargs):
        ...



from attr import define, field
import tiktoken
from skatepark.utils import Tokenizer


@define(frozen=True)
class TiktokenTokenizer(Tokenizer):
    DEFAULT_MODEL = "gpt-3.5-turbo"
    DEFAULT_ENCODING = "cl100k_base"
    DEFAULT_MAX_TOKENS = 2049
    TOKEN_OFFSET = 8

    MODEL_PREFIXES_TO_MAX_TOKENS = {
        "gpt-4-32k": 32768,
        "gpt-4": 8192,
        "gpt-3.5-turbo": 4096,
        "text-davinci-003": 4097,
        "text-davinci-002": 4097,
        "code-davinci-002": 8001
    }

    CHAT_API_PREFIXES = [
        "gpt-3.5-turbo",
        "gpt-4"
    ]

    model: str = field(default=DEFAULT_MODEL, kw_only=True)
    stop_sequence: str = field(default=Tokenizer.DEFAULT_STOP_SEQUENCE, kw_only=True)

    def encode(self, text: str) -> list[int]:
        return self.encoding.encode(text, allowed_special={self.stop_sequence})

    def decode(self, tokens: list[int]) -> str:
        return self.encoding.decode(tokens)

    def token_count(self, text: str) -> int:
        return len(self.encode(text))

    def tokens_left(self, text: str) -> int:
        diff = self.max_tokens - self.token_count(text)

        if diff > 0:
            return diff
        else:
            return 0

    def is_chat(self) -> bool:
        return next(p for p in self.CHAT_API_PREFIXES if self.model.startswith(p)) is not None

    @property
    def encoding(self) -> tiktoken.Encoding:
        try:
            return tiktoken.encoding_for_model(self.model)
        except KeyError:
            return tiktoken.get_encoding(self.DEFAULT_ENCODING)

    @property
    def max_tokens(self) -> int:
        tokens = next(v for k, v in self.MODEL_PREFIXES_TO_MAX_TOKENS.items() if self.model.startswith(k))

        return (tokens if tokens else self.DEFAULT_MAX_TOKENS) - self.TOKEN_OFFSET



import json
import openai
from attr import define, field, Factory
from skatepark.drivers import PromptDriver
from skatepark.utils import TiktokenTokenizer
from skatepark.artifacts import TextOutput


@define
class OpenAiPromptDriver(PromptDriver):
    model: str = field(default=TiktokenTokenizer.DEFAULT_MODEL, kw_only=True)
    tokenizer: TiktokenTokenizer = field(
        default=Factory(lambda self: TiktokenTokenizer(model=self.model), takes_self=True),
        kw_only=True
    )
    temperature: float = field(default=0.5, kw_only=True)
    user: str = field(default="", kw_only=True)

    def try_run(self, value: any) -> TextOutput:
        if self.tokenizer.is_chat():
            return self.__run_chat(value)
        else:
            return self.__run_completion(value)

    def __run_chat(self, value: str) -> TextOutput:
        result = openai.ChatCompletion.create(
            model=self.tokenizer.model,
            messages=[
                {
                    "role": "user",
                    "content": value
                }
            ],
            max_tokens=self.tokenizer.tokens_left(value),
            temperature=self.temperature,
            stop=self.tokenizer.stop_sequence,
            user=self.user
        )

        if len(result.choices) == 1:
            return TextOutput(
                value=result.choices[0]["message"]["content"].strip(),
                meta={
                    "id": result["id"],
                    "created": result["created"],
                    "usage": json.dumps(result["usage"])
                }
            )
        else:
            raise Exception("Completion with more than one choice is not supported yet.")

    def __run_completion(self, value: str) -> TextOutput:
        result = openai.Completion.create(
            model=self.tokenizer.model,
            prompt=value,
            max_tokens=self.tokenizer.tokens_left(value),
            temperature=self.temperature,
            stop=self.tokenizer.stop_sequence,
            user=self.user
        )

        if len(result.choices) == 1:
            return TextOutput(
                value=result.choices[0].text.strip(),
                meta={
                    "id": result["id"],
                    "created": result["created"],
                    "usage": json.dumps(result["usage"])
                }
            )
        else:
            raise Exception("Completion with more than one choice is not supported yet.")


import uuid

from attr import define, field, Factory

from skatepark.utils import J2


@define
class PipelineRun:
    id: str = field(default=Factory(lambda: uuid.uuid4().hex), kw_only=True)
    input: str = field(kw_only=True)
    output: str = field(kw_only=True)

    def render(self) -> str:
        return J2("prompts/run_context.j2").render(
            run=self
        )



import json
from skatepark.utils.tokenizer import Tokenizer
from skatepark.utils.tiktoken_tokenizer import TiktokenTokenizer
from skatepark.utils.j2 import J2
from skatepark.utils.conversation import Conversation
from skatepark.utils.tool_loader import ToolLoader

__all__ = [
    "Tokenizer",
    "TiktokenTokenizer",
    "J2",
    "Conversation",
    "ToolLoader"
]


def minify_json(value: str) -> str:
    return json.dumps(json.loads(value), separators=(',', ':'))



from abc import abstractmethod
from pydoc import locate
from typing import Optional
from marshmallow import ValidationError, Schema
from skatepark.schemas import BaseSchema


class PolymorphicSchema(BaseSchema):
    """
    PolymorphicSchema is based on https://github.com/marshmallow-code/marshmallow-oneofschema
    """

    def get_schema(self, class_name: str, obj: Optional[object], schema_namespace: Optional[str]):
        if schema_namespace:
            namespace = schema_namespace
        elif obj is not None and hasattr(obj, "schema_namespace"):
            if locate(f"skatepark.schemas.{class_name}Schema"):
                namespace = "skatepark.schemas"
            elif obj.schema_namespace is None:
                namespace = obj.schema_namespace = f"{obj.__module__}_schema"
            else:
                namespace = obj.schema_namespace
        else:
            namespace = "skatepark.schemas"

        klass = locate(f"{namespace}.{class_name}Schema")

        if klass:
            return klass
        else:
            raise ValidationError(f"Missing schema for '{class_name}'")

    type_field = "type"
    type_field_remove = True

    def get_obj_type(self, obj):
        """Returns name of the schema during dump() calls, given the object
        being dumped."""
        return obj.__class__.__name__

    def get_data_type(self, data):
        """Returns name of the schema during load() calls, given the data being
        loaded. Defaults to looking up `type_field` in the data."""
        data_type = data.get(self.type_field)
        if self.type_field in data and self.type_field_remove:
            data.pop(self.type_field)
        return data_type

    def dump(self, obj, *, many=None, **kwargs):
        errors = {}
        result_data = []
        result_errors = {}
        many = self.many if many is None else bool(many)
        if not many:
            result = result_data = self._dump(obj, **kwargs)
        else:
            for idx, o in enumerate(obj):
                try:
                    result = self._dump(o, **kwargs)
                    result_data.append(result)
                except ValidationError as error:
                    result_errors[idx] = error.normalized_messages()
                    result_data.append(error.valid_data)

        result = result_data
        errors = result_errors

        if not errors:
            return result
        else:
            exc = ValidationError(errors, data=obj, valid_data=result)
            raise exc

    def _dump(self, obj, *, update_fields=True, **kwargs):
        obj_type = self.get_obj_type(obj)

        if not obj_type:
            return (
                None,
                {"_schema": "Unknown object class: %s" % obj.__class__.__name__},
            )

        type_schema = self.get_schema(obj_type, obj, None)

        if not type_schema:
            return None, {"_schema": "Unsupported object type: %s" % obj_type}

        schema = type_schema if isinstance(type_schema, Schema) else type_schema()

        schema.context.update(getattr(self, "context", {}))

        result = schema.dump(obj, many=False, **kwargs)

        if result is not None:
            result[self.type_field] = obj_type

        return result

    def load(self, data, *, many=None, partial=None, unknown=None, **kwargs):
        errors = {}
        result_data = []
        result_errors = {}
        many = self.many if many is None else bool(many)
        if partial is None:
            partial = self.partial
        if not many:
            try:
                result = result_data = self._load(
                    data, partial=partial, unknown=unknown, **kwargs
                )
                #  result_data.append(result)
            except ValidationError as error:
                result_errors = error.normalized_messages()
                result_data.append(error.valid_data)
        else:
            for idx, item in enumerate(data):
                try:
                    result = self._load(item, partial=partial, **kwargs)
                    result_data.append(result)
                except ValidationError as error:
                    result_errors[idx] = error.normalized_messages()
                    result_data.append(error.valid_data)

        result = result_data
        errors = result_errors

        if not errors:
            return result
        else:
            exc = ValidationError(errors, data=data, valid_data=result)
            raise exc

    def _load(self, data, *, partial=None, unknown=None, **kwargs):
        if not isinstance(data, dict):
            raise ValidationError({"_schema": "Invalid data type: %s" % data})

        data = dict(data)
        unknown = unknown or self.unknown
        data_type = self.get_data_type(data)

        if data_type is None:
            raise ValidationError(
                {self.type_field: ["Missing data for required field."]}
            )

        schema_namespace = data.get("schema_namespace")

        try:
            type_schema = self.get_schema(data_type, None, schema_namespace)
        except TypeError:
            # data_type could be unhashable
            raise ValidationError({self.type_field: ["Invalid value: %s" % data_type]})
        if not type_schema:
            raise ValidationError(
                {self.type_field: ["Unsupported value: %s" % data_type]}
            )

        schema = type_schema if isinstance(type_schema, Schema) else type_schema()

        schema.context.update(getattr(self, "context", {}))

        return schema.load(data, many=False, partial=partial, unknown=unknown, **kwargs)

    def validate(self, data, *, many=None, partial=None):
        try:
            self.load(data, many=many, partial=partial)
        except ValidationError as ve:
            return ve.messages
        return {}

    @abstractmethod
    def make_obj(self, data, **kwargs):
        ...



from abc import abstractmethod
from marshmallow import fields
from skatepark.schemas import BaseSchema


class SummarizerSchema(BaseSchema):
    class Meta:
        ordered = True

    type = fields.Str(required=True)

    @abstractmethod
    def make_obj(self, data, **kwargs):
        ...



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from abc import ABC, abstractmethod
from attr import define, field, Factory

if TYPE_CHECKING:
    from skatepark.memory import PipelineMemory, PipelineRun


@define
class Summarizer(ABC):
    type: str = field(default=Factory(lambda self: self.__class__.__name__, takes_self=True), kw_only=True)

    @abstractmethod
    def summarize(self, memory: PipelineMemory, runs: list[PipelineRun]) -> Optional[str]:
        ...



from abc import ABC, abstractmethod
from skatepark.memory import PipelineMemory


class MemoryDriver(ABC):
    @abstractmethod
    def store(self, memory: PipelineMemory) -> None:
        ...

    @abstractmethod
    def load(self) -> PipelineMemory:
        ...



from marshmallow import post_load
from skatepark.schemas import StructureSchema


class WorkflowSchema(StructureSchema):
    @post_load
    def make_obj(self, data, **kwargs):
        from skatepark.structures import Workflow

        return Workflow(**data)



from typing import Optional

from attr import define, field
from griptape.core import BaseTool, BaseExecutor
from griptape.core.executors import LocalExecutor


@define
class ToolLoader:
    tools: list[BaseTool] = field(factory=list, kw_only=True)
    executor: BaseExecutor = field(default=LocalExecutor(), kw_only=True)

    def load_tool(self, tool_name: str) -> Optional[BaseTool]:
        return next((t for t in self.tools if t.name == tool_name), None)



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from attr import define, field
from skatepark.utils import J2
from skatepark.drivers import PromptDriver
from skatepark.summarizers.summarizer import Summarizer


if TYPE_CHECKING:
    from skatepark.memory import PipelineMemory, PipelineRun


@define
class PromptDriverSummarizer(Summarizer):
    driver: PromptDriver = field(kw_only=True)

    def summarize(self, memory: PipelineMemory, runs: list[PipelineRun]) -> Optional[str]:
        try:
            if len(runs) > 0:
                return self.driver.run(
                    value=J2("prompts/summarize.j2").render(
                        summary=memory.summary,
                        runs=runs
                    )
                ).value
            else:
                return memory.summary
        except Exception as e:
            self.pipeline.logger.error(f"Error summarizing memory: {type(e).__name__}({e})")

            return memory.summary



from attr import define, field
from skatepark.drivers import MemoryDriver
from skatepark.memory import PipelineMemory


@define
class DiskMemoryDriver(MemoryDriver):
    file_path: str = field(default="skatepark_memory.json", kw_only=True)

    def store(self, memory: PipelineMemory) -> None:
        with open(self.file_path, "w") as file:
            file.write(memory.to_json())

    def load(self) -> PipelineMemory:
        with open(self.file_path, "r") as file:
            memory = PipelineMemory.from_json(file.read())

            memory.driver = self

            return memory



from marshmallow import fields, post_load
from skatepark.schemas import BaseSchema, PipelineRunSchema, PipelineMemorySchema


class BufferPipelineMemorySchema(PipelineMemorySchema):
    buffer_size = fields.Int()

    @post_load
    def make_obj(self, data, **kwargs):
        from skatepark.memory import BufferPipelineMemory

        return BufferPipelineMemory(**data)



from abc import ABC, abstractmethod
from typing import Optional


class Tokenizer(ABC):
    DEFAULT_STOP_SEQUENCE = "Observation:"

    model: str
    stop_sequence: str

    @abstractmethod
    def encode(self, text: str) -> list[int]:
        ...

    @abstractmethod
    def decode(self, tokens: list[int]) -> str:
        ...

    @abstractmethod
    def token_count(self, text: str) -> int:
        ...

    @abstractmethod
    def tokens_left(self, text: str) -> Optional[int]:
        ...

    @abstractmethod
    def encoding(self):
        ...

    @abstractmethod
    def max_tokens(self) -> Optional[int]:
        ...



from __future__ import annotations
import concurrent.futures as futures
import json
from graphlib import TopologicalSorter
from attr import define, field
from skatepark.artifacts import ErrorOutput
from skatepark.steps import Step
from skatepark.structures import Structure
from skatepark.utils import J2


@define
class Workflow(Structure):
    executor: futures.Executor = field(default=futures.ThreadPoolExecutor(), kw_only=True)

    def add_step(self, step: Step) -> Step:
        step.structure = self

        self.steps.append(step)

        return step

    def prompt_stack(self, step: Step) -> list[str]:
        stack = Structure.prompt_stack(self, step)

        stack.append(
            J2("prompts/workflow.j2").render(
                step=step
            )
        )

        return stack

    def run(self, *args) -> list[Step]:
        self._execution_args = args
        ordered_steps = self.order_steps()
        exit_loop = False

        while not self.is_finished() and not exit_loop:
            futures_list = {}

            for step in ordered_steps:
                if step.can_execute():
                    future = self.executor.submit(step.execute)
                    futures_list[future] = step

            # Wait for all tasks to complete
            for future in futures.as_completed(futures_list):
                if isinstance(future.result(), ErrorOutput):
                    exit_loop = True

                    break

        self._execution_args = ()

        return self.output_steps()

    def context(self, step: Step) -> dict[str, any]:
        context = super().context(step)

        context.update(
            {
                "inputs": {parent.id: parent.output.value if parent.output else "" for parent in step.parents},
                "parents": {parent.id: parent for parent in step.parents},
                "children": {child.id: child for child in step.children}
            }
        )

        return context

    def output_steps(self) -> list[Step]:
        return [step for step in self.steps if not step.children]

    def to_graph(self) -> dict[str, set[str]]:
        graph: dict[str, set[str]] = {}

        for key_step in self.steps:
            graph[key_step.id] = set()

            for value_step in self.steps:
                if key_step.id in value_step.child_ids:
                    graph[key_step.id].add(value_step.id)

        return graph

    def order_steps(self) -> list[Step]:
        return [self.find_step(step_id) for step_id in TopologicalSorter(self.to_graph()).static_order()]

    def to_dict(self) -> dict:
        from skatepark.schemas import WorkflowSchema

        return WorkflowSchema().dump(self)

    @classmethod
    def from_dict(cls, workflow_dict: dict) -> Workflow:
        from skatepark.schemas import WorkflowSchema

        return WorkflowSchema().load(workflow_dict)

    @classmethod
    def from_json(cls, workflow_json: str) -> Workflow:
        return Workflow.from_dict(json.loads(workflow_json))



from marshmallow import post_load, fields
from skatepark.schemas import StructureSchema


class PipelineSchema(StructureSchema):
    autoprune_memory = fields.Bool()

    @post_load
    def make_obj(self, data, **kwargs):
        from skatepark.structures import Pipeline

        return Pipeline(**data)



from skatepark.memory.pipeline_run import PipelineRun
from skatepark.memory.pipeline_memory import PipelineMemory
from skatepark.memory.summary_pipeline_memory import SummaryPipelineMemory
from skatepark.memory.buffer_pipeline_memory import BufferPipelineMemory


__all__ = [
    "PipelineRun",
    "PipelineMemory",
    "SummaryPipelineMemory",
    "BufferPipelineMemory"
]



from marshmallow import fields, post_load
from skatepark.schemas import PipelineMemorySchema, PolymorphicSchema


class SummaryPipelineMemorySchema(PipelineMemorySchema):
    offset = fields.Int()
    summary = fields.Str()
    summary_index = fields.Int()
    summarizer = fields.Nested(PolymorphicSchema())

    @post_load
    def make_obj(self, data, **kwargs):
        from skatepark.memory import SummaryPipelineMemory

        return SummaryPipelineMemory(**data)



from abc import abstractmethod
from marshmallow import fields
from skatepark.schemas import BaseSchema, PolymorphicSchema


class PromptDriverSchema(BaseSchema):
    class Meta:
        ordered = True

    max_retries = fields.Int()
    retry_delay = fields.Float()
    model = fields.Str()
    tokenizer = fields.Nested(PolymorphicSchema())

    @abstractmethod
    def make_obj(self, data, **kwargs):
        ...



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from attr import define, field
from skatepark.utils import J2
from skatepark.steps import Step
from skatepark.artifacts import TextOutput

if TYPE_CHECKING:
    from skatepark.drivers import PromptDriver


@define
class PromptStep(Step):
    prompt_template: str = field(default="{{ args[0] }}")
    context: dict[str, any] = field(factory=dict, kw_only=True)
    driver: Optional[PromptDriver] = field(default=None, kw_only=True)

    def before_run(self) -> None:
        super().before_run()

        self.structure.logger.info(f"Step {self.id}\nInput: {self.render_prompt()}")

    def run(self) -> TextOutput:
        self.output = self.active_driver().run(value=self.structure.to_prompt_string(self))

        return self.output

    def after_run(self) -> None:
        super().after_run()

        self.structure.logger.info(f"Step {self.id}\nOutput: {self.output.value}")

    def active_driver(self) -> PromptDriver:
        if self.driver is None:
            return self.structure.prompt_driver
        else:
            return self.driver

    def render_prompt(self) -> str:
        return J2().render_from_string(
            self.prompt_template,
            **self.full_context
        )

    def render(self) -> str:
        return J2("prompts/steps/prompt.j2").render(
            step=self
        )

    @property
    def full_context(self) -> dict[str, any]:
        structure_context = self.structure.context(self)

        structure_context.update(self.context)

        return structure_context



from marshmallow import fields, post_load
from skatepark.schemas import BaseSchema


class RuleSchema(BaseSchema):
    value = fields.Str()

    @post_load
    def make_obj(self, data, **kwargs):
        from skatepark.rules import Rule

        return Rule(**data)



from __future__ import annotations
import uuid
from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Optional
from attr import define, field, Factory
from skatepark.artifacts import ErrorOutput

if TYPE_CHECKING:
    from skatepark.artifacts import TextOutput, StructureArtifact
    from skatepark.steps import Step
    from skatepark.structures import Structure


@define
class Step(ABC):
    class State(Enum):
        PENDING = 1
        EXECUTING = 2
        FINISHED = 3

    id: str = field(default=Factory(lambda: uuid.uuid4().hex), kw_only=True)
    state: State = field(default=State.PENDING, kw_only=True)
    type: str = field(default=Factory(lambda self: self.__class__.__name__, takes_self=True), kw_only=True)
    parent_ids: list[str] = field(factory=list, kw_only=True)
    child_ids: list[str] = field(factory=list, kw_only=True)

    output: Optional[StructureArtifact] = field(default=None, init=False)
    structure: Optional[Structure] = field(default=None, init=False)

    @property
    def parents(self) -> list[Step]:
        return [self.structure.find_step(parent_id) for parent_id in self.parent_ids]

    @property
    def children(self) -> list[Step]:
        return [self.structure.find_step(child_id) for child_id in self.child_ids]

    def add_child(self, child: Step) -> Step:
        if self.structure:
            child.structure = self.structure
        elif child.structure:
            self.structure = child.structure

        if child not in self.structure.steps:
            self.structure.steps.append(child)

        if self not in self.structure.steps:
            self.structure.steps.append(self)

        if child.id not in self.child_ids:
            self.child_ids.append(child.id)

        if self.id not in child.parent_ids:
            child.parent_ids.append(self.id)

        return child

    def add_parent(self, parent: Step) -> Step:
        if self.structure:
            parent.structure = self.structure
        elif parent.structure:
            self.structure = parent.structure

        if parent not in self.structure.steps:
            self.structure.steps.append(parent)

        if self not in self.structure.steps:
            self.structure.steps.append(self)

        if parent.id not in self.parent_ids:
            self.parent_ids.append(parent.id)

        if self.id not in parent.child_ids:
            parent.child_ids.append(self.id)

        return parent

    def is_pending(self) -> bool:
        return self.state == Step.State.PENDING

    def is_finished(self) -> bool:
        return self.state == Step.State.FINISHED

    def is_executing(self) -> bool:
        return self.state == Step.State.EXECUTING

    def before_run(self) -> None:
        pass

    def after_run(self) -> None:
        pass

    def execute(self) -> StructureArtifact:
        try:
            self.state = Step.State.EXECUTING

            self.before_run()

            self.output = self.run()

            self.after_run()
        except Exception as e:
            self.structure.logger.error(f"Step {self.id}\n{e}", exc_info=True)

            self.output = ErrorOutput(str(e), exception=e, step=self)
        finally:
            self.state = Step.State.FINISHED

            return self.output

    def can_execute(self) -> bool:
        return self.state == Step.State.PENDING and all(parent.is_finished() for parent in self.parents)

    def reset(self) -> Step:
        self.state = Step.State.PENDING
        self.output = None

        return self

    @abstractmethod
    def run(self) -> TextOutput:
        ...



from marshmallow import fields, post_load
from skatepark.schemas import PromptDriverSchema


class OpenAiPromptDriverSchema(PromptDriverSchema):
    temperature = fields.Float()
    user = fields.Str()

    @post_load
    def make_obj(self, data, **kwargs):
        from skatepark.drivers import OpenAiPromptDriver

        return OpenAiPromptDriver(**data)



from marshmallow import fields
from skatepark.schemas import PolymorphicSchema, SummarizerSchema


class PromptDriverSummarizerSchema(SummarizerSchema):
    driver = fields.Nested(PolymorphicSchema())

    def make_obj(self, data, **kwargs):
        from skatepark.summarizers import PromptDriverSummarizer

        return PromptDriverSummarizer(**data)
